/***************************************************************************
 * aiff.c: Virtual AIFF file access routines.
 * Copyright (C) 2011 Michael C McTernan,
 *    Michael.McTernan.2001@cs.bris.ac.uk
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ***************************************************************************/

/**************************************************************************
 * Includes
 **************************************************************************/

#include <pthread.h>
#include <arpa/inet.h>
#include <endian.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include "FLAC/stream_decoder.h"
#include "aiff.h"
#include "id3.h"

/**************************************************************************
 * Manfest Constants
 **************************************************************************/

#define AIFF_ID3_ALLOC_CHUNK 1024

#define AIFF_FLAC_NOMINAL_BLOCK_SIZE 4096

/**************************************************************************
 * Macros
 **************************************************************************/

#define M_Min(a, b)  ((a) > (b) ? (b) : (a))
#define M_Max(a, b)  ((a) < (b) ? (b) : (a))

/**************************************************************************
 * Types
 **************************************************************************/

/** Structure to hold a block of decoded data.
 * This is used to cache decoded data to give better performance on small
 * writes or seeks.
 */
typedef struct aiff_decoded_block
{
    /** Offset within aiff file at which this block would appear */
    off_t   byteOffset;

    /** Size of the block in bytes */
    size_t  size;

    /** Size of this whole structure. */
    size_t  allocSize;

    /** Actual data */
    uint8_t data[];
}
aiff_decoded_block_t;


/** Type for a four character code ID field in AIFF files.
 */
typedef struct  __attribute__((packed))
{
    char field[4];
}
aiff_id_t;


/** Data related to a 'virtual' AIFF file.
 * This stores the file chunks that would be present in the AIFF file as well
 * as context for the decoder and some book keeping variables.
 */
struct aiff
{
    struct __attribute__((packed))
    {
        struct __attribute__((packed))
        {
            aiff_id_t     ckId;     /* 'FORM' */
            uint32_t      ckSize;   /* Total size */
            aiff_id_t     formType; /* 'AIFF' */
        }
        hdr;

        struct __attribute__((packed))
        {
            aiff_id_t       ckId;            /* 'COMM' */
            int32_t         ckSize;          /* 18 */
            int16_t         numChannels;
            uint32_t        numSampleFrames;
            int16_t         sampleSize;
            char            sampleRate[10];  /* 80-bit float */
        }
        commonChunk;

        struct __attribute__((packed))
        {
            aiff_id_t       ckId;      /* 'SSND' */
            int32_t         ckSize;
            uint32_t        offset;    /* 0 */
            uint32_t        blockSize; /* 0 */
        }
        soundDataChunk;
    }
    hd;

    struct __attribute__((packed))
    {
        aiff_id_t       ckId;      /* 'ID3 ' */
        int32_t         ckSize;
    }
    id3Chunk;

    /** Lock to serialise access to decoder and cache. */
    pthread_mutex_t       lock;

    /** The ID3 tag data, if present, otherwise NULL. */
    id3tag_t              id3tag;

    /** Decoder context. */
    FLAC__StreamDecoder  *flacDecoder;

    /** Current request for decoding data */
    void                 *reqBuf;
    off_t                 reqOffset;
    size_t                reqCount;
    size_t                reqReadCount;

    /** Last decoded block. */
    aiff_decoded_block_t *decBlock;

    /** The next sample to be decoded. */
    uint64_t              decNextSample;

    /** Stats on the decoded block cache. */
    uint32_t              hitCount, missCount;

    /** Count of seeks within the FLAC file. */
    uint32_t              seekCount;
};

/**************************************************************************
 * Local Variables
 **************************************************************************/

/**************************************************************************
 * Local Functions
 **************************************************************************/

/** Set an Id field.
 */
static void setId(aiff_id_t *id, const char *data)
{
    assert(strlen(data) == 4);

    for(uint8_t t = 0; t < 4; t++)
    {
        id->field[t] = data[t];
    }
}


/** Try to satisfy a request with the passed data buffer.
 */
static bool addToReqBuffer(struct aiff *a, const void *buf, size_t count, off_t offset)
{
    const uint8_t *bbuf = buf;

    /* Check for overlap at the start of the buffer */
    if(offset <= a->reqOffset && offset + (off_t)count > a->reqOffset)
    {
        size_t bufOffset = a->reqOffset - offset;
        size_t cpyLen = M_Min(a->reqCount, count - bufOffset);

        memcpy(a->reqBuf, bbuf + bufOffset, cpyLen);
        a->reqBuf += cpyLen;
        a->reqOffset += cpyLen;
        a->reqCount -= cpyLen;
        a->reqReadCount += cpyLen;

        return true;
    }

    return false;
}


static void setStreamInfo(struct aiff *a, const FLAC__StreamMetadata_StreamInfo *info)
{
    union
    {
        long double ld;
        char        c[12];
    }
    u;
    uint32_t    size;

    a->hd.soundDataChunk.offset = 0;
    a->hd.soundDataChunk.blockSize = 0;

    a->hd.commonChunk.ckSize = htonl(18);
    a->hd.commonChunk.numChannels = htons(info->channels);
    a->hd.commonChunk.numSampleFrames = htonl(info->total_samples);
    a->hd.commonChunk.sampleSize = htons(info->bits_per_sample);

    /* Copy the 80-bit IEEE 754 float */
    u.ld = info->sample_rate;
    for(uint8_t t = 0; t < 10; t++)
    {
        a->hd.commonChunk.sampleRate[9 - t] = u.c[t];
    }

    /* Add the size required for the sample data */
    size = info->channels * info->total_samples * (info->bits_per_sample / 8);

    /* Ths ssnd chunk size is the audio data + chunk header */
    a->hd.soundDataChunk.ckSize = htonl(size + 8);
}


static void addMetaComment(struct aiff *a, const FLAC__StreamMetadata_VorbisComment *info)
{
    /* Check if memory needs allocating */
    if(!a->id3tag)
    {
        a->id3tag = Id3New();
    }

    /* Check each comment in turn */
    for(uint32_t c = 0; c < info->num_comments; c++)
    {
        Id3AddFlacFrame(a->id3tag,(char *)info->comments[c].entry);
    }

    /* Check if the tag is empty */
    if(Id3IsEmpty(a->id3tag))
    {
        Id3Free(a->id3tag);
        a->id3tag = NULL;
    }
    else
    {
        a->id3Chunk.ckSize = htonl(Id3Length(a->id3tag));
    }
}


void addMetaPicture(struct aiff *a, const FLAC__StreamMetadata_Picture *pic)
{
    /* Check if memory needs allocating */
    if(!a->id3tag)
    {
        a->id3tag = Id3New();
    }

    /* Add the picture */
    Id3AddFlacPicture(a->id3tag, pic->type, pic->mime_type, (char *)pic->description, pic->data_length, pic->data);

    /* Check if the tag is empty */
    if(Id3IsEmpty(a->id3tag))
    {
        Id3Free(a->id3tag);
        a->id3tag = NULL;
    }
    else
    {
        a->id3Chunk.ckSize = htonl(Id3Length(a->id3tag));
    }
}


FLAC__StreamDecoderWriteStatus callbackWrite(const FLAC__StreamDecoder *decoder __attribute__((unused)),
                                             const FLAC__Frame         *frame,
                                             const FLAC__int32   *const buffer[],
                                             void                      *data)
{
    struct aiff   *a = (struct aiff *)data;
    const uint8_t  bitsPerSample  = frame->header.bits_per_sample;
    const uint8_t  bytesPerSample = (bitsPerSample + 7) / 8;
    uint8_t       *buf, *bufEnd;

    assert(frame->header.number_type == FLAC__FRAME_NUMBER_TYPE_SAMPLE_NUMBER);

    /* Record the last sample that has been decoded */
    a->decNextSample = frame->header.blocksize +
                       frame->header.number.sample_number;

    /* Determine storage for decoded data */
    size_t decBlockSize = sizeof(aiff_decoded_block_t) +
                                 (frame->header.blocksize * frame->header.channels * bytesPerSample);

    /* Ensure there is enough memory for the decode block */
    if(__builtin_expect(a->decBlock == NULL || a->decBlock->allocSize < decBlockSize, 0))
    {
        if(a->decBlock)
        {
            free(a->decBlock);
        }

        if(decBlockSize < AIFF_FLAC_NOMINAL_BLOCK_SIZE)
        {
            decBlockSize = AIFF_FLAC_NOMINAL_BLOCK_SIZE;
        }

        a->decBlock = malloc(decBlockSize);
        if(!a->decBlock)
        {
            return FLAC__STREAM_DECODER_WRITE_STATUS_ABORT;
        }
        a->decBlock->allocSize = decBlockSize;
    }

    /* Determine the byte offset in the aiff output file where the first
     *  decoded sample would appear.
     */
    a->decBlock->byteOffset = sizeof(a->hd) +
                              (frame->header.number.sample_number * bytesPerSample * frame->header.channels);

    bufEnd = buf = a->decBlock->data;

    /* Check for the common case of 16-bit stereo data on little-endian */
    if(__builtin_expect(bytesPerSample == 2 && frame->header.channels == 2 &&
                        frame->header.blocksize == AIFF_FLAC_NOMINAL_BLOCK_SIZE &&
                        BYTE_ORDER == LITTLE_ENDIAN, 1))
    {
        for(uint_fast32_t s = 0; s < AIFF_FLAC_NOMINAL_BLOCK_SIZE; s++)
        {
            for(uint_fast8_t c = 0; c < 2; c++)
            {
                *bufEnd = (buffer[c][s] >> 8) & 0xff;
                bufEnd++;
                *bufEnd = (buffer[c][s] >> 0) & 0xff;
                bufEnd++;
            }
        }
    }
    else /* Generic implementation */
    {
        for(uint32_t s = 0; s < frame->header.blocksize; s++)
        {
            for(uint8_t c = 0; c < frame->header.channels; c++)
            {
                union
                {
                    uint32_t sample;
                    char     c[4];
                }
                u;

                /* Change to network byte order */
                u.sample = htonl(buffer[c][s]);

                /* Write required bytes */
                memcpy(bufEnd, &u.c[4 - bytesPerSample], bytesPerSample);
                bufEnd += bytesPerSample;
            }
        }
    }

    a->decBlock->size = bufEnd - buf;

    addToReqBuffer(a, buf, bufEnd - buf, a->decBlock->byteOffset);

    return FLAC__STREAM_DECODER_WRITE_STATUS_CONTINUE;
}


void callbackMetadata(const FLAC__StreamDecoder  *decoder __attribute__((unused)),
                      const FLAC__StreamMetadata *metadata,
                      void                       *data)
{
    struct aiff *aiff = (struct aiff *)data;

    switch(metadata->type)
    {
        case FLAC__METADATA_TYPE_STREAMINFO:
            setStreamInfo(aiff, &metadata->data.stream_info);
            break;
        case FLAC__METADATA_TYPE_VORBIS_COMMENT:
            addMetaComment(aiff, &metadata->data.vorbis_comment);
            break;
        case FLAC__METADATA_TYPE_PICTURE:
            addMetaPicture(aiff, &metadata->data.picture);
            break;
        case FLAC__METADATA_TYPE_SEEKTABLE:
            break;
        default:
            break;
    }
}


void callbackError(const FLAC__StreamDecoder      *decoder __attribute__((unused)),
                   FLAC__StreamDecoderErrorStatus  status,
                   void                           *data __attribute__((unused)))
{
    //struct aiff *aiff = (struct aiff *)data;

    fprintf(stderr, "Aifffffs: Got error callback: %s\n", FLAC__StreamDecoderErrorStatusString[status]);
}

/**************************************************************************
 * Global Functions
 **************************************************************************/

struct aiff *AiffNew(FILE *f, bool disableMd5Verify)
{
    FLAC__StreamDecoderInitStatus status;
    struct aiff *a;

    /* Try to allocate memory */
    a = calloc(sizeof(struct aiff), 1);
    if(!a)
    {
        goto fail;
    }

    /* Create the lock */
    pthread_mutex_init(&a->lock, NULL);

    /* Fill the fields that can be determined at this point */
    setId(&a->hd.hdr.ckId, "FORM");
    setId(&a->hd.hdr.formType, "AIFF");
    setId(&a->hd.commonChunk.ckId, "COMM");
    setId(&a->hd.soundDataChunk.ckId, "SSND");
    setId(&a->id3Chunk.ckId, "ID3 ");

    /* Create the decoder */
    a->flacDecoder = FLAC__stream_decoder_new();
    if(a->flacDecoder == NULL)
    {
        goto fail_lock;
    }

    /* Configure the decoder */
    FLAC__stream_decoder_set_md5_checking(a->flacDecoder, !disableMd5Verify);
    FLAC__stream_decoder_set_metadata_respond_all(a->flacDecoder);

    status = FLAC__stream_decoder_init_FILE(a->flacDecoder, f, callbackWrite,
                                            callbackMetadata, callbackError, a);
    if(status != FLAC__STREAM_DECODER_INIT_STATUS_OK)
    {
        goto fail_delete;
    }

    /* Decode all meta-data */
    if(!FLAC__stream_decoder_process_until_end_of_metadata(a->flacDecoder))
    {
        goto fail_delete;
    }

    uint32_t hdrSize;

    /* Compute size to put in the header */
    hdrSize = 4 + ntohl(a->hd.commonChunk.ckSize) + 8 +
                  ntohl(a->hd.soundDataChunk.ckSize) + 8;

    /* Add ID3 block if it has been added */
    if(a->id3tag)
    {
        hdrSize += ntohl(a->id3Chunk.ckSize) + 8;
    }

    a->hd.hdr.ckSize = htonl(hdrSize);

    return a;

fail_delete:
    if(a->id3tag)
    {
        Id3Free(a->id3tag);
    }

    FLAC__stream_decoder_delete(a->flacDecoder);
fail_lock:
    pthread_mutex_destroy(&a->lock);
    free(a);
fail:
    return NULL;
}


size_t AiffSize(struct aiff *a)
{
    size_t s = sizeof(a->hd) + ntohl(a->hd.soundDataChunk.ckSize) - 8;

    if(a->id3tag)
    {
        s += Id3Length(a->id3tag) + sizeof(a->id3Chunk);
    }

    return s;
}


size_t AiffRead(struct aiff *a, void *buf, size_t count, off_t offset)
{
    const uint8_t  channels = ntohs(a->hd.commonChunk.numChannels);
    const uint16_t bytesPerSample = (ntohs(a->hd.commonChunk.sampleSize) + 7) / 8;
    size_t         result;

    pthread_mutex_lock(&a->lock);

    /* Current request for decoding data */
    a->reqBuf = buf;
    a->reqOffset = offset;
    a->reqCount = count;
    a->reqReadCount = 0;

    /* Try adding data from the header */
    addToReqBuffer(a, &a->hd, sizeof(a->hd), 0);

    /* Check if a previous decode can satisfy the request */
    if(a->decBlock != NULL && a->reqCount != 0)
    {
        if(addToReqBuffer(a, a->decBlock->data, a->decBlock->size, a->decBlock->byteOffset))
        {
            a->hitCount++;
        }
        else
        {
            a->missCount++;
        }
    }

    /* Compute offsets for the audio data */
    const off_t adStart = sizeof(a->hd);
    const off_t adEnd = sizeof(a->hd) + ntohl(a->hd.soundDataChunk.ckSize) - 8;
    bool        needSeek = true;

    /* Check if we need to decode samples */
    while(a->reqOffset >= adStart && a->reqOffset < adEnd && a->reqCount != 0)
    {
        /* Check if we need to seek to the required sample position */
        if(needSeek)
        {
            FLAC__uint64 sample = (a->reqOffset - adStart) / (bytesPerSample * channels);

            assert(sample <= FLAC__stream_decoder_get_total_samples(a->flacDecoder));

            /* Check if the decoder is already at the correct position */
            if(sample == a->decNextSample)
            {
                needSeek = false;
            }

            /* Check if seek is needed */
            if(needSeek)
            {
                /* Attempt to seek */
                if(!FLAC__stream_decoder_seek_absolute(a->flacDecoder, sample))
                {
                    pthread_mutex_unlock(&a->lock);
                    return -EIO;
                }

                /* Furthed decodes will be sequential for this block */
                needSeek = false;
                a->seekCount++;
            }
        }

        if(!FLAC__stream_decoder_process_single(a->flacDecoder))
        {
            pthread_mutex_unlock(&a->lock);
            return -EIO;
        }
    }

    /* Check the offsets for ID3 headers if present */
    if(a->id3tag != NULL && a->reqCount != 0)
    {
        addToReqBuffer(a, &a->id3Chunk, sizeof(a->id3Chunk), adEnd);
        addToReqBuffer(a, Id3Data(a->id3tag), Id3Length(a->id3tag), adEnd + sizeof(a->id3Chunk));
    }

    result = a->reqReadCount;

    pthread_mutex_unlock(&a->lock);

    return result;
}


bool AiffDestroy(struct aiff *a)
{
    bool streamOk;

    printf("%" PRIu32 " / %" PRIu32 " = %2.5f, %" PRIu32 " seeks\n",
           a->hitCount, a->missCount +  a->hitCount,
           (float)a->hitCount / (float)(a->missCount +  a->hitCount),
           a->seekCount);

    /* Finish the decoder and check if the MD5 failed */
    streamOk = FLAC__stream_decoder_finish(a->flacDecoder);

    /* Delete the decoder */
    FLAC__stream_decoder_delete(a->flacDecoder);

    /* Free the last decode block */
    if(a->decBlock)
    {
        free(a->decBlock);
    }

    /* Free ID3 tag allocations if present */
    if(a->id3tag)
    {
        Id3Free(a->id3tag);
    }

    pthread_mutex_destroy(&a->lock);

    free(a);

    return streamOk;
}


/**************************************************************************
 * Global Functions
 **************************************************************************/

#ifdef MODULE_TEST

int main(int argc, char *argv[])
{
    struct aiff *af;
    FILE *fin, *fout;
    size_t s;
    off_t  offset;

    fin = fopen(argv[1], "rb");
    fout = fopen(argv[2], "wb");

    af = AiffNew(fin, true);

    s = AiffSize(af);
    printf("Reported size: %u\n", s);

    offset = 0;
    do
    {
        char buf[131072];

        s = AiffRead(af, buf, sizeof(buf), offset);
        fwrite(buf, s, 1, fout);
        offset += s;
    }
    while(s == 131072);

    printf("Read size: %u\n", offset);

    AiffDestroy(af);
    fclose(fout);

    return EXIT_SUCCESS;
}

#endif

/* END OF FILE */

