/***************************************************************************
 * aifffffs.h: AIFF from FLAC file system.
 * Copyright (C) 2011 Michael C McTernan,
 *    Michael.McTernan.2001@cs.bris.ac.uk
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ***************************************************************************/

/**************************************************************************
 * Includes
 **************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <fuse.h>
#include <ulockmgr.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <fcntl.h>
#include <stdio.h>
#include <dirent.h>
#include <errno.h>
#include <assert.h>
#include <sys/time.h>
#ifdef HAVE_SETXATTR
#include <sys/xattr.h>
#endif
#include <mcheck.h>
#include "aiff.h"
#include "aifffffs.h"
#include "aifffffs_dir.h"
#include "aifffffs_file.h"

/**************************************************************************
 * Manfest Constants
 **************************************************************************/

/**************************************************************************
 * Macros
 **************************************************************************/

/**************************************************************************
 * Types
 **************************************************************************/

/** Command line option key values.
 */
typedef enum
{
    OPT_FLAC_ROOT_SHORT,
    OPT_FLAC_ROOT_LONG,
    OPT_FLAC_MD5_SHORT,
    OPT_FLAC_MD5_LONG,
}
aifffffs_opt_t;

/**************************************************************************
 * Local Variables
 **************************************************************************/

/** List of permitted command line options.
 */
static struct fuse_opt aifffffs_opts[] = {
    FUSE_OPT_KEY("-r ",            OPT_FLAC_ROOT_SHORT),
    FUSE_OPT_KEY("--root ",        OPT_FLAC_ROOT_LONG),
    FUSE_OPT_KEY("-m",             OPT_FLAC_MD5_SHORT),
    FUSE_OPT_KEY("--no-md5",       OPT_FLAC_MD5_LONG),
    FUSE_OPT_END
};

/**************************************************************************
 * Global Variables
 **************************************************************************/

/** Filesystem root and the length of the string. */
char    *aifffffs_root = NULL;
uint32_t aifffffs_rootLen = 0;

/** Disable MD5 validation on FLAC files. */
bool     aifffffs_noMd5 = false;

/**************************************************************************
 * Filesystem Functions
 **************************************************************************/

static int Aifffffs_getattr(const char *path, struct stat *stbuf)
{
    const char *fixedPath = M_UnTranslatePath(path);

    /* Check if this is for an aiff file */
    if(isAiffFilename(path))
    {
        if(lstat(fixedPath, stbuf) == -1)
        {
            return -errno;
        }

        /* Now modify the length to reflect decoded verion */
        FILE *flac = fopen(fixedPath, "rb");
        if(flac == NULL)
        {
            return -errno;
        }

        /* Open the file, disabling MD5 checking */
        struct aiff *af = AiffNew(flac, true);

        stbuf->st_size = AiffSize(af);

        /* Fix the block count so programs like du get the correct size.
         *  man 2 stat indicates that st_blocks is in 512 byte blocks.
         */
        stbuf->st_blocks = (stbuf->st_size + 511) / 512;

        AiffDestroy(af);

        return 0;
    }
    else if(lstat(fixedPath, stbuf) == -1)
    {
        return -errno;
    }
    else
    {
        return 0;
    }
}


static int Aifffffs_access(const char *path, int mask)
{
    const char *rpath = M_UnTranslatePath(path);

    if(access(rpath, mask) == -1)
        return -errno;
    else
        return 0;
}


static int Aifffffs_readlink(const char *path, char *buf, size_t size)
{
    const char *rpath = M_UnTranslatePath(path);
    ssize_t len;

    len = readlink(rpath, buf, size - 1);
    if(len == -1)
    {
        return -errno;
    }

    /* readlink() doesn't null terminate */
    buf[len] = '\0';
    return 0;
}


static int Aifffffs_mknod(const char *path, mode_t mode, dev_t rdev)
{
    int res;

    if(S_ISFIFO(mode))
        res = mkfifo(path, mode);
    else
        res = mknod(path, mode, rdev);

    if(res == -1)
        return -errno;

    return 0;
}


static int Aifffffs_mkdir(const char *path, mode_t mode)
{
    if(mkdir(path, mode) == -1)
        return -errno;
    else
        return 0;
}


static int Aifffffs_unlink(const char *path)
{
    const char *rpath = M_UnTranslatePath(path);

    if(unlink(rpath) == -1)
        return -errno;
    else
        return 0;
}


static int Aifffffs_rmdir(const char *path)
{
    if(rmdir(path) == -1)
        return -errno;
    else
        return 0;
}


static int Aifffffs_symlink(const char *from, const char *to)
{
    const char *rfrom = M_UnTranslatePath(from);
    const char *rto   = M_UnTranslatePath(to);

    if(symlink(rfrom, rto) == -1)
        return -errno;
    else
        return 0;
}


static int Aifffffs_rename(const char *from, const char *to)
{
    const char *rfrom = M_UnTranslatePath(from);
    const char *rto   = M_UnTranslatePath(to);

    if(rename(rfrom, rto) == -1)
        return -errno;
    else
        return 0;
}


static int Aifffffs_link(const char *from, const char *to)
{
    const char *rfrom = M_UnTranslatePath(from);
    const char *rto   = M_UnTranslatePath(to);

    if(link(rfrom, rto) == -1)
        return -errno;
    else
        return 0;
}


static int Aifffffs_chmod(const char *path, mode_t mode)
{
    const char *rpath = M_UnTranslatePath(path);

    if(chmod(rpath, mode) == -1)
        return -errno;
    else
        return 0;
}


static int Aifffffs_chown(const char *path, uid_t uid, gid_t gid)
{
    const char *rpath = M_UnTranslatePath(path);

    if(lchown(rpath, uid, gid) == -1)
        return -errno;
    else
        return 0;
}


static int Aifffffs_utimens(const char *path, const struct timespec ts[2])
{
    const char *rpath = M_UnTranslatePath(path);
    struct timeval tv[2];

    tv[0].tv_sec = ts[0].tv_sec;
    tv[0].tv_usec = ts[0].tv_nsec / 1000;
    tv[1].tv_sec = ts[1].tv_sec;
    tv[1].tv_usec = ts[1].tv_nsec / 1000;

    if(utimes(rpath, tv) == -1)
        return -errno;
    else
        return 0;
}


int Aifffffs_truncate(const char *path, off_t size)
{
    const char *rpath = M_UnTranslatePath(path);

    if(truncate(rpath, size) == -1)
        return -errno;
    else
        return 0;
}


static int Aifffffs_statfs(const char *path, struct statvfs *stbuf)
{
    int res;

    res = statvfs(path, stbuf);
    if (res == -1)
        return -errno;

    return 0;
}


static int Aifffffs_fsync(const char *path, int isdatasync,
             struct fuse_file_info *fi)
{
    int res;
    (void) path;

#ifndef HAVE_FDATASYNC
    (void) isdatasync;
#else
    if (isdatasync)
        res = fdatasync(fi->fh);
    else
#endif
        res = fsync(fi->fh);
    if (res == -1)
        return -errno;

    return 0;
}


#ifdef HAVE_SETXATTR
/* xattr operations are optional and can safely be left unimplemented */
static int Aifffffs_getxattr(const char *path, const char *name, char *value,
            size_t size)
{
    int res = lgetxattr(path, name, value, size);
    if (res == -1)
        return -errno;
    return res;
}

static int Aifffffs_listxattr(const char *path, char *list, size_t size)
{
    int res = llistxattr(path, list, size);
    if (res == -1)
        return -errno;
    return res;
}
#endif /* HAVE_SETXATTR */


static struct fuse_operations Aifffffs_oper = {
    .getattr     = Aifffffs_getattr,
    .fgetattr    = Aifffffs_fgetattr,
    .access      = Aifffffs_access,
    .readlink    = Aifffffs_readlink,
    .opendir     = Aifffffs_opendir,
    .readdir     = Aifffffs_readdir,
    .releasedir  = Aifffffs_releasedir,
    .mknod       = Aifffffs_mknod,
    .mkdir       = Aifffffs_mkdir,
    .symlink     = Aifffffs_symlink,
    .unlink      = Aifffffs_unlink,
    .rmdir       = Aifffffs_rmdir,
    .rename      = Aifffffs_rename,
    .link        = Aifffffs_link,
    .chmod       = Aifffffs_chmod,
    .chown       = Aifffffs_chown,
    .truncate    = Aifffffs_truncate,
    .ftruncate   = Aifffffs_ftruncate,
    .utimens     = Aifffffs_utimens,
    .create      = Aifffffs_create,
    .open        = Aifffffs_open,
    .read        = Aifffffs_read,
    .write       = Aifffffs_write,
    .statfs      = Aifffffs_statfs,
    .flush       = Aifffffs_flush,
    .release     = Aifffffs_release,
    .fsync       = Aifffffs_fsync,
#ifdef HAVE_SETXATTR
    .getxattr    = Aifffffs_getxattr,
    .listxattr   = Aifffffs_listxattr,
#endif

  /*.flag_nullpath_ok = 0,*/
};

/**************************************************************************
 * Local Functions
 **************************************************************************/

/** Callback to parse options either from the aifffffs_opts list or unknown.
 */
static int Aifffffs_ParseOpts(void             *data __attribute__((unused)),
                              const char       *arg,
                              int               key,
                              struct fuse_args *outArgs __attribute__((unused)))
{
    switch(key)
    {
        case OPT_FLAC_ROOT_SHORT:
            /* Hack of leading -r and duplicate */
            aifffffs_root = realpath(&arg[2], NULL);
            aifffffs_rootLen = aifffffs_root ? strlen(aifffffs_root) : 0;

            /* Now discard the option */
            return 0;

        case OPT_FLAC_ROOT_LONG:
            /* Hack of leading -root and duplicate */
            aifffffs_root = realpath(&arg[5], NULL);
            aifffffs_rootLen = aifffffs_root ? strlen(aifffffs_root) : 0;

            /* Now discard the option */
            return 0;

        case OPT_FLAC_MD5_SHORT:
        case OPT_FLAC_MD5_LONG:
            aifffffs_noMd5 = true;
            return 0;

        default:
            return 1; /* keep the option */
    }
}

/**************************************************************************
 * Global Functions
 **************************************************************************/

int main(int argc, char *argv[])
{
    struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
    struct stat      sbuf;

    /* Start memory tracer if needed */
    if(getenv("MALLOC_TRACE"))
    {
        fprintf(stderr, "Notice: Malloc tracing enabled.\n");
        mtrace();
    }

    fuse_opt_parse(&args, NULL, aifffffs_opts, Aifffffs_ParseOpts);

    if(aifffffs_root == NULL)
    {
        fprintf(stderr, "Error: FLAC directory root for filesystem not specified or invalid.\n");
        return EXIT_FAILURE;
    }
    else if(stat(aifffffs_root, &sbuf) != 0)
    {
        fprintf(stderr, "Error: FLAC directory root for filesystem invalid: '%s': %m\n", aifffffs_root);
        return EXIT_FAILURE;
    }
    else if(!S_ISDIR(sbuf.st_mode))
    {
        fprintf(stderr, "Error: FLAC directory root for filesystem invalid: '%s': not a directory\n", aifffffs_root);
        return EXIT_FAILURE;
    }

    return fuse_main(args.argc, args.argv, &Aifffffs_oper, NULL);
}

/* END OF FILE */
