/***************************************************************************
 * aifffffs_dir.c: Directory handling.
 * Copyright (C) 2011 Michael C McTernan,
 *    Michael.McTernan.2001@cs.bris.ac.uk
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ***************************************************************************/

/**************************************************************************
 * Includes
 **************************************************************************/

#include <sys/types.h>
#include <stdbool.h>
#include <dirent.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include "aifffffs.h"
#include "aifffffs_dir.h"

/**************************************************************************
 * Manfest Constants
 **************************************************************************/

/**************************************************************************
 * Macros
 **************************************************************************/

/**************************************************************************
 * Types
 **************************************************************************/

/**************************************************************************
 * Local Variables
 **************************************************************************/

/**************************************************************************
 * Local Functions
 **************************************************************************/

static bool mangleName(char *n, const struct stat *statbuf)
{
    /* Check if it's a regular file */
    if(S_ISREG(statbuf->st_mode))
    {
        uint32_t len = strlen(n);

        /* Check the extension to see if it need renaming */
        if(len > 5 && strcasecmp(".flac", &n[len - 5]) == 0)
        {
            strcpy(&n[len - 5], ".aiff");
            return true;
        }
    }
    else if(S_ISDIR(statbuf->st_mode))
    {
        /* Directories are always allowed */
        return true;
    }

    return false;
}

/**************************************************************************
 * Global Functions
 **************************************************************************/

int Aifffffs_opendir(const char *path, struct fuse_file_info *fi)
{
    size_t      pathLen = strlen(path);
    char *const rpath = alloca(pathLen + aifffffs_rootLen);
    DIR        *d;

    /* Adjust to be the relative to the aifffffs root */
    strcpy(rpath, aifffffs_root);
    strcpy(&rpath[aifffffs_rootLen], path);

    /* Open the directory */
    d = opendir(rpath);
    if(d == NULL)
    {
        return -errno;
    }

    /* Store the the directory pointer structure */
    fi->fh = (unsigned long)d;

    return 0;
}


int Aifffffs_readdir(const char            *path,
                     void                  *buf,
                     fuse_fill_dir_t        filler,
                     off_t                  offset,
                     struct fuse_file_info *fi)
{
    DIR           *d = (DIR *)(uintptr_t)fi->fh;
    uint32_t       pathLen = strlen(path);
    struct dirent *de;
    struct dirent  deBuf;
    struct stat    statbuf;
    bool           skip;

    /* Seek to the requested offset */
    seekdir(d, offset);

    do
    {
        do
        {
            char fullname[PATH_MAX];

            /* Read the directory entry */
            readdir_r(d, &deBuf, &de);
            if(de == NULL)
            {
                /* End of the directory has been reached */
                return 0;
            }

            /* Construct the full path */
            if(aifffffs_rootLen + pathLen + strlen(de->d_name) + 2 < PATH_MAX)
            {
                strcpy(fullname, aifffffs_root);
                strcat(fullname, path);
                strcat(fullname, "/");
                strcat(fullname, de->d_name);

                /* Populate the stats */
                skip = (stat(fullname, &statbuf) != 0);
            }
            else
            {
                skip = 1;
            }
        }
        while(skip || !mangleName(de->d_name, &statbuf));
    }
    while(!filler(buf, de->d_name, &statbuf, telldir(d)));

    return 0;
}


int Aifffffs_releasedir(const char            *path __attribute__((unused)),
                        struct fuse_file_info *fi)
{
    DIR *d = (DIR *)(uintptr_t)fi->fh;

    closedir(d);

    return 0;
}

/* END OF FILE */
