/***************************************************************************
 * aifffffs_file.c: Virtual file handling.
 * Copyright (C) 2011 Michael C McTernan,
 *    Michael.McTernan.2001@cs.bris.ac.uk
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ***************************************************************************/

/**************************************************************************
 * Includes
 **************************************************************************/

#include <fuse.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdint.h>
#include <errno.h>
#include <string.h>
#include <assert.h>
#include "aiff.h"
#include "aifffffs.h"
#include "aifffffs_file.h"

/**************************************************************************
 * Manfest Constants
 **************************************************************************/

/**************************************************************************
 * Macros
 **************************************************************************/

/**************************************************************************
 * Types
 **************************************************************************/

/**************************************************************************
 * Local Variables
 **************************************************************************/

/**************************************************************************
 * Local Functions
 **************************************************************************/

/**************************************************************************
 * Global Functions
 **************************************************************************/

int Aifffffs_open(const char *path, struct fuse_file_info *fi)
{
    const char *rpath = M_UnTranslatePath(path);

    if(!isAiffFilename(path) || (fi->flags & (O_WRONLY | O_RDWR)) != 0)
    {
        return -EACCES;
    }

    FILE *f = fopen(rpath, "rb");
    if(f == NULL)
    {
        return -errno;
    }

    fi->fh = (unsigned long)AiffNew(f, aifffffs_noMd5);
    if(fi->fh == 0)
        return -ENOMEM;
    else
        return 0;
}


int Aifffffs_create(const char            *path __attribute__((unused)),
                    mode_t                 mode __attribute__((unused)),
                    struct fuse_file_info *fi   __attribute__((unused)))
{
    return -EACCES;
}


int Aifffffs_ftruncate(const char            *path __attribute__((unused)),
                       off_t                  size __attribute__((unused)),
                       struct fuse_file_info *fi   __attribute__((unused)))
{
    return -EACCES;
}


int Aifffffs_read(const char            *path __attribute__((unused)),
                  char                  *buf,
                  size_t                 size,
                  off_t                  offset,
                  struct fuse_file_info *fi)
{
    struct aiff *f = (struct aiff *)(uintptr_t)fi->fh;

    return AiffRead(f, buf, size, offset);
}


int Aifffffs_write(const char            *path   __attribute__((unused)),
                   const char            *buf    __attribute__((unused)),
                   size_t                 size   __attribute__((unused)),
                   off_t                  offset __attribute__((unused)),
                   struct fuse_file_info *fi     __attribute__((unused)))
{
    return -EACCES;
}


int Aifffffs_flush(const char            *path __attribute__((unused)),
                   struct fuse_file_info *fi __attribute__((unused)))
{
    /* This is called from every close on an open file, so call the
       close on the underlying filesystem.  But since flush may be
       called multiple times for an open file, this must not really
       close the file.  This is important if used on a network
       filesystem like NFS which flush the data/metadata on close() */
    //res = close(dup(fi->fh));
    //if (res == -1)
    //    return -errno;

    return 0;
}


int Aifffffs_fgetattr(const char            *path __attribute__((unused)),
                      struct stat           *stbuf,
                      struct fuse_file_info *fi)
{
    const char  *fixedPath = M_UnTranslatePath(path);

    /* Check if this is for an aiff file */
    if(isAiffFilename(path))
    {
        if(lstat(fixedPath, stbuf) == -1)
        {
            return -errno;
        }

        /* Now modify the length to reflect decoded verion */
        struct aiff *af = (struct aiff *)(uintptr_t)fi->fh;

        assert(af != NULL);

        stbuf->st_size = AiffSize(af);

        return 0;
    }
    else if(lstat(fixedPath, stbuf) == -1)
    {
        return -errno;
    }
    else
    {
        return 0;
    }
}


int Aifffffs_release(const char            *path __attribute__((unused)),
                     struct fuse_file_info *fi)
{
    struct aiff *f = (struct aiff *)(uintptr_t)fi->fh;

    /* Destroy the stream and check it was okay */
    if(!AiffDestroy(f))
    {
        return -EIO;
    }

    return 0;
}

/* END OF FILE */
