/***************************************************************************
 * id3.c: ID3 tag construction
 * Copyright (C) 2011 Michael C McTernan,
 *    Michael.McTernan.2001@cs.bris.ac.uk
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ***************************************************************************/

/**************************************************************************
 * Includes
 **************************************************************************/

#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <stdio.h>
#include <iconv.h>

/**************************************************************************
 * Manfest Constants
 **************************************************************************/

/**************************************************************************
 * Macros
 **************************************************************************/

#define TAG_BUFFER_BYTES 4096

#define TAG_USE_UTF8        0

/**************************************************************************
 * Types
 **************************************************************************/

typedef struct
{
    char     tag[4];
    uint32_t len;
    char     data[];
}
id3_payload_t;

struct id3tag
{
    /** Pointer to the datablock. */
    uint8_t  *data;

    /** Length of the datablock. */
    uint32_t  len;

    /** Size of the datablock in terms of memory allocation. */
    uint32_t  allocSize;
};

/**************************************************************************
 * Local Variables
 **************************************************************************/

static const uint8_t id3hdr[] = { 'I', 'D', '3',
                                  0x03, 0x00, /* Version */
                                  0x00,
                                  0x00, 0x00, 0x00, 0x00 /* Size */};

/**************************************************************************
 * Local Functions
 **************************************************************************/

static bool ensureCapacity(struct id3tag *id3, uint32_t size)
{
    if(id3->allocSize < size)
    {
        id3->allocSize = size;
        id3->data = realloc(id3->data, size);
    }

    return id3->data != NULL;
}


static uint32_t codeString(char *stringUtf8, char *outdata, uint32_t outdataLen)
{
    size_t inbytes = strlen(stringUtf8);
    size_t otbytes = outdataLen - 1;
    char  *in = stringUtf8, *out = &outdata[1];

    /* Use UTF-16 encoding */
    outdata[0] = 1;

    iconv_t ic = iconv_open("UTF-16", "UTF-8");
    if(ic == (iconv_t)-1)
    {
        fprintf(stderr, "Error: Could not open iconv for UTF-8 -> UTF-16 transcoding\n");
        return 0;
    }

    iconv(ic, &in, &inbytes, &out, &otbytes);
    iconv_close(ic);

    *out = '\0';
    out++;
    *out = '\0';
    out++;

    return out - outdata;
}


/** Add a tag.
 */
uint32_t addTag(struct id3tag *id3, id3_payload_t *payload)
{
    ensureCapacity(id3, id3->len + payload->len + 10);

    /* Frame header, 4 bytes */
    memcpy(&id3->data[id3->len], payload->tag, 4);
    id3->len += 4;

    /* Size, 4 bytes */
    id3->data[id3->len++] = (payload->len) >> 24;
    id3->data[id3->len++] = (payload->len) >> 16;
    id3->data[id3->len++] = (payload->len) >>  8;
    id3->data[id3->len++] = (payload->len) >>  0;

    /* Flags, 2 bytes */
    id3->data[id3->len++] = 0;
    id3->data[id3->len++] = 0;

    /* Add the data */
    memcpy(&id3->data[id3->len], payload->data, payload->len);
    id3->len += payload->len;

    /* Update the ID3 tag length */
    uint32_t tagLen = id3->len - sizeof(id3hdr);

    id3->data[6] = (tagLen >> 21) & 0x7f;
    id3->data[7] = (tagLen >> 14) & 0x7f;
    id3->data[8] = (tagLen >>  7) & 0x7f;
    id3->data[9] = (tagLen >>  0) & 0x7f;

    return tagLen;
}


/** Extract the MusicBrainz ID from some URL.
 */
static char *extractMbid(char *mbUrl)
{
    uint32_t t = strlen(mbUrl);

    while(t > 0 && mbUrl[t - 1] != '/')
    {
        t--;
    }

    return &mbUrl[t];
}


static bool addFlacSimple(struct id3tag *id3, char *flacFrame)
{
    /* Simple/straight map, see:
     *  http://wiki.musicbrainz.org/PicardTagMapping
     *  http://wiki.musicbrainz.org/MusicBrainz_Tag
     */
    static const struct { char *flac, *id3; } smap[] =
    {
        { "GROUPING=",             "TIT1" },
        { "TITLE=",                "TIT2" },
        { "TITLESORT=",            "TSOT" },
        { "SUBTITLE=",             "TIT3" },
        { "GENRE=",                "TCON" },
        { "MOOD=",                 "TMOD" },
        { "ISRC=",                 "TSRC" },
        { "MEDIA=",                "TMED" },
        { "LABEL=",                "TPUB" },
        { "DISCSUBTITLE=",         "TSST" },
        { "TRACKNUMBER=",          "TRCK" },
        { "DISCNUMBER=",           "TPOS" },
        { "ALBUM=",                "TALB" },
        { "ALBUMSORT=",            "TSOA" },
        { "ARTIST=",               "TPE1" },
        { "ARTISTSORT=",           "TSOP" },
        { "ALBUMARTIST=",          "TPE2" },
        { "ALBUMARTISTSORT=",      "TSO2" },
        { "CONDUCTOR=",            "TPE3" },
        { "COMPOSERSORT=",         "TSOC" },
        { "REMIXER=",              "TPE4" },
        { "PERFORMER=",            "TPE1" },
        { "COMPILATION=",          "TCMP" },
    };

    for(uint16_t t = 0; t < sizeof(smap) / sizeof(smap[0]); t++)
    {
        uint16_t flacLen = strlen(smap[t].flac);

        if(strncmp(smap[t].flac, flacFrame, flacLen) == 0)
        {
            id3_payload_t *pl = alloca(TAG_BUFFER_BYTES);

            memcpy(pl->tag, smap[t].id3, 4);

            if(TAG_USE_UTF8)
            {
                /* Straight UTF-8, which isn't allowed in ID3v2.3 */
                pl->data[0] = 0;
                strcpy(&pl->data[1], &flacFrame[flacLen]);

                pl->len = strlen(&pl->data[1]) + 2;
            }
            else
            {
                /* Use UTF-16 encoding */
                pl->len = codeString(&flacFrame[flacLen], pl->data,
                                     TAG_BUFFER_BYTES - offsetof(id3_payload_t, data));
            }

            addTag(id3, pl);

            return true;
        }
    }

    return false;
}


static bool addFlacTxxx(struct id3tag *id3, char *flacFrame)
{
    /* TXXX map */
    static const struct { char *flac, *desc; bool cutMbId; } txxxMap[] =
    {
        { "MUSICBRAINZ_ARTISTID=",      "MusicBrainz Artist Id",       true  },
        { "MUSICBRAINZ_ALBUMID=",       "MusicBrainz Album Id",        true  },
        { "MUSICBRAINZ_ALBUMARTISTID=", "MusicBrainz Album Artist Id", true  },
        { "MUSICBRAINZ_TRMID=",         "MusicBrainz TRM Id",          false },
        { "ASIN=",                      "ASIN",                        false },
    };

    for(uint16_t t = 0; t < sizeof(txxxMap) / sizeof(txxxMap[0]); t++)
    {
        uint16_t flacLen = strlen(txxxMap[t].flac);

        if(strncmp(txxxMap[t].flac, flacFrame, flacLen) == 0)
        {
            id3_payload_t *pl = alloca(TAG_BUFFER_BYTES);
            char          *val;

            /* Get the value to use for the tag */
            val = &flacFrame[flacLen];
            if(txxxMap[t].cutMbId)
            {
                val = extractMbid(val);
            }

            memcpy(pl->tag, "TXXX", 4);

            /* Encoding */
            pl->data[0] = 0;
            pl->len = 1;

            /* Null terminated description */
            strcpy(&pl->data[pl->len], txxxMap[t].desc);
            pl->len += strlen(txxxMap[t].desc) + 1;

            /* Binary data; in this case I make it null terminated too */
            strcpy(&pl->data[pl->len], val);
            pl->len += strlen(val) + 1;

            addTag(id3, pl);

            return true;
        }
    }

    return false;
}

/**************************************************************************
 * Global Functions
 **************************************************************************/

struct id3tag *Id3New(void)
{
    struct id3tag *id3 = calloc(sizeof(struct id3tag), 1);

    if(!id3)
    {
        return NULL;
    }

    if(ensureCapacity(id3, sizeof(id3hdr)))
    {
        memcpy(id3->data, id3hdr, sizeof(id3hdr));
        id3->len = sizeof(id3hdr);

        return id3;
    }
    else
    {
        free(id3);
        return NULL;
    }
}


/** Check if the tag is empty.
 */
bool Id3IsEmpty(struct id3tag *id3)
{
    return id3->len == sizeof(id3hdr);
}


/** Add a FLAC frame to the tag.
 *
 * \param[in] id3        ID3 tag to which the frame should be added.
 * \param[in] flacFrame  The whole FLAC tag, UTF-8 encoded as per convention.
 */
void Id3AddFlacFrame(struct id3tag *id3, char *flacFrame)
{
    if(addFlacSimple(id3, flacFrame))
    {
    }
    else if(addFlacTxxx(id3, flacFrame))
    {
    }
    else if(strncmp("MUSICBRAINZ_TRACKID=", flacFrame, 20) == 0)
    {
        id3_payload_t *pl = alloca(TAG_BUFFER_BYTES);
        char          *mbid = extractMbid(&flacFrame[20]);

        memcpy(pl->tag, "UFID", 4);

        strcpy(pl->data, "http://musicbrainz.org");
        pl->len = strlen(pl->data) + 1;

        strcpy(&pl->data[pl->len], mbid);
        pl->len += strlen(mbid) + 1;

        addTag(id3, pl);
    }
    else if(strncmp("RELEASE=GB, ", flacFrame, 12) == 0)
    {
        id3_payload_t *pl = alloca(TAG_BUFFER_BYTES);
        uint16_t       flacLen = 12;

        memcpy(pl->tag, "TYER", 4);

        flacFrame[flacLen + 4] = '\0';

        if(TAG_USE_UTF8)
        {
            /* Straight UTF-8, which isn't allowed in ID3v2.3 */
            pl->data[0] = 0;
            strcpy(&pl->data[1], &flacFrame[flacLen]);
            pl->len = strlen(&flacFrame[flacLen]) + 2;
        }
        else
        {
            /* Use UTF-16 encoding */
            pl->len = codeString(&flacFrame[flacLen], pl->data, TAG_BUFFER_BYTES - offsetof(id3_payload_t, data));
        }

        addTag(id3, pl);


        if(strlen(&flacFrame[flacLen + 5]) >= 5)
        {
            char date[5];

            memcpy(pl->tag, "TDAT", 4);

            date[0] = flacFrame[flacLen + 8];
            date[1] = flacFrame[flacLen + 9];
            date[2] = flacFrame[flacLen + 5];
            date[3] = flacFrame[flacLen + 6];
            date[4] = '\0';

            /* Use UTF-16 encoding */
            pl->len = codeString(date, pl->data, TAG_BUFFER_BYTES - offsetof(id3_payload_t, data));

            addTag(id3, pl);
        }
    }
    else
    {
        fprintf(stderr, "Warning: Dropping FLAC tag %s\n", flacFrame);
    }
}


void Id3AddFlacPicture(struct id3tag *id3,
                       uint32_t       type,
                       char          *mimeType,
                       char          *description,
                       uint32_t       dataLen,
                       void          *data)
{
    id3_payload_t *pl = alloca(dataLen + TAG_BUFFER_BYTES);
    uint32_t       typeByte;

    memcpy(pl->tag, "APIC", 4);

    /* Set UTF-16 encoding */
    pl->data[0] = 1;

    /* Set the mime type and null terminate */
    strcpy(&pl->data[1], mimeType);
    pl->len = strlen(mimeType) + 2;

    /* Note the next byte index which holds the image type */
    typeByte = pl->len;

    /* Add the encoded description */
    pl->len += codeString(description, &pl->data[pl->len], dataLen + TAG_BUFFER_BYTES - offsetof(id3_payload_t, data));

    /* Overwrite what would be the encoding type added by codeString())
     *  with the image type.
     */
    pl->data[typeByte] = type;

    memcpy(&pl->data[pl->len], data, dataLen);
    pl->len += dataLen;

    addTag(id3, pl);
}


void *Id3Data(struct id3tag *id3)
{
    return id3->data;
}


uint32_t Id3Length(struct id3tag *id3)
{
    return id3->len;
}


void Id3Free(struct id3tag *id3)
{
    if(id3->data)
    {
        free(id3->data);
    }
    free(id3);
}

/* END OF FILE */


